#include "littlelib.h"
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#define MAX 20
static uint16_t devlist[MAX];
static little_t device;
static int quiet = 0;

static void myshutdown(int error) {
    if (device) {
        little_close(device);
    }
    little_shutdown();
    exit(error);
}

static void usage(char *name) {
    printf("Usage: %s [--host <host>] [--port <port>][--device <n>] [list | reset | get | poll | on | off | ledon | ledoff | scratch]\n", name);
}

static void list() {
    int count = little_list(devlist, MAX);
    for (int i=0;i<count;i++) {
        little_t v = little_open(devlist[i]);
        if (v) {
            uint8_t val = little_get(v);
            printf("LittleUSB %d: input=%d output=%d led=%d\n", i, (val&1) ? 1 : 0, (val&2) ? 1 : 0, (val&4) ? 1 : 0);
            little_close(v);
        }
    }
}

static void scratch(char *host, int port) {
    uint8_t state = 0;
    const uint8_t BUFLEN = 64;
    unsigned char buf[BUFLEN+1];
    char *c;
    int olstate = 0;
    fd_set rs;
    struct timeval timeout;
    int l = 0, n = 0, lstate = 0;
    uint32_t len = 0;
    int fd;
    struct sockaddr_in serv_addr;
    struct hostent *server;

    while (1) {

        if (!device) {
            int count = little_list(devlist, MAX);
            if (count > 0) {
                device = little_open(devlist[0]);
                if (device && !quiet) {
                    printf("Scratch: opened device %04x\n", devlist[0]);
                }
            }
        }
        if (device) {
            if (state == 0) {
                fd = socket(AF_INET, SOCK_STREAM, 0);
                if (fd < 0) {
                    if (!quiet) {
                        perror("opening socket");
                    }
                    return;
                }
                if (!(server = gethostbyname(host))) {
                    if (!quiet) {
                        perror("can't find host");
                    }
                } else {
                    bzero((char *) &serv_addr, sizeof(serv_addr));
                    serv_addr.sin_family = AF_INET;
                    bcopy((char *)server->h_addr, (char *)&serv_addr.sin_addr.s_addr, server->h_length);
                    serv_addr.sin_port = htons(port);
                    if (connect(fd, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) < 0) {
                        if (!quiet) {
                            perror("connect");
                        }
                        state = 0;
                    } else {
                        if (!quiet) {
                            printf("Scratch: connected to %s:%d\n", host, port);
                        }
                        state = 1;
                        c = (char *)buf;
                        len = 0;
                        n = 4;
                    }
                }
            }

            if (state == 1) {
                lstate = little_get(device);
                if (lstate < 0) {
                    printf("Scratch: error reading device %04x, resetting\n", devlist[0]);
                    little_free(device);
                    device = NULL;
                    state = 2;
                }
            }

            if (state == 1) {
                timeout.tv_sec = 0;
                timeout.tv_usec = 250000;
                FD_ZERO(&rs);
                FD_SET(fd, &rs);
                if ((l=select(fd+1, &rs, NULL, NULL, &timeout)) > 0 && FD_ISSET(fd, &rs)) {
                    if ((l=read(fd, c, 1)) < 0) {
                        if (!quiet) {
                            perror("socket read");
                        }
                        state = 2;
                    } else if (l > 0) {
                        c++;
                        n--;
                        if (n == 0) {
                            if (len == 0) {
                                len = (buf[0]<<24) | (buf[1]<<16) | (buf[2]<<8) | (buf[3]);
                                if (len <= BUFLEN-4) {
                                    n = len;
                                    c = (void*)buf;
                                } else {
                                    if (!quiet) {
                                        fprintf(stderr, "Scratch: invalid len %d\n", len);
                                    }
                                    state = 2;
                                }
                            } else {
                                *c = 0;
                                if (!quiet) {
                                    printf("Scratch: read %s\n", buf);
                                }
                                char *d = (void*)buf;
                                if ((c = strsep(&d, " "))) {
                                    if (!strcmp(c, "broadcast")) {
                                        if ((c = strsep(&d, " "))) {
                                            if (!strcmp(c, "\"littlebit-output-on\"")) {
                                                lstate |= 2;
                                            } else if (!strcmp(c, "\"littlebit-output-off\"")) {
                                                lstate &= ~2;
                                            } else if (!strcmp(c, "\"littlebit-output-flip\"")) {
                                                lstate = (lstate & 2) ==0 ? lstate | 2 : lstate & ~2;
                                            } else if (!strcmp(c, "\"littlebit-led-on\"")) {
                                                lstate |= 4;
                                            } else if (!strcmp(c, "\"littlebit-led-off\"")) {
                                                lstate &= ~4;
                                            } else if (!strcmp(c, "\"littlebit-led-flip\"")) {
                                                lstate = (lstate & 4) == 0 ? lstate | 4 : lstate & ~4;
                                            } else {
                                                continue;
                                            }
                                            if (little_set(device, lstate) < 0) {
                                                printf("Scratch: error reading device %04x, resetting\n", devlist[0]);
                                                little_free(device);
                                                device = NULL;
                                                state = 2;
                                            }
                                        }
                                    }
                                }
                                n = 4;
                                len = 0;
                                c = (void*)buf;
                            }
                        }
                    }
                } else if (l < 0) {
                    if (!quiet) {
                        perror("select");
                    }
                    state = 2;
                }
            }

            if (state == 1) {
                char *resp = NULL;
                if ((lstate&1) & (!(olstate&1))) {
                    resp = "broadcast \"littlebit-input-on\"";
                } else if ((!(lstate&1)) & (olstate&1)) {
                    resp = "broadcast \"littlebit-input-off\"";
                }
                if (resp) {
                    if (!quiet) {
                        printf("Scratch: write %s", resp);
                    }
                    uint32_t size = htons(strlen(resp));
                    write(fd, &size, 4);
                    write(fd, resp, strlen(resp));
                }
                olstate = lstate;
            }

            if (state == 2) {
                printf("Closing\n");
                close(fd);
                fd = 0;
                state = 0;
            }

            if (state == 0) {
                sleep(1);
            }
        }
    }
}

static little_t open(int num) {
    int count = little_list(devlist, MAX);
    little_t device = NULL;
    if (count <= num) {
        if (count == 0) {
            fprintf(stderr, "No devices\n");
        } else {
            fprintf(stderr, "No device %d\n", num);
        }
        myshutdown(-1);
    } else {
        device = little_open(devlist[num]);
        if (!device) {
            fprintf(stderr, "Can't open %d\n", num);
            myshutdown(-1);
        }
    }
    return device;
}

int main(int argc, char **argv) {
    little_init();
    if (argc == 1) {
        usage(argv[0]);
        myshutdown(-1);
    }
    int devicenum = 0;
    int error = 0, state = 0;
    int port = 42001;
    char *host = "127.0.0.1";
    struct sigaction sigact;
    sigset_t mask, oldmask;

    memset(&sigact, 0, sizeof(sigact));
    sigact.sa_handler = &myshutdown;
    sigaction(SIGINT, &sigact, 0);
    sigaction(SIGTERM, &sigact, 0);
    sigaction(SIGHUP, &sigact, 0);
    sigemptyset(&mask);
    sigprocmask(SIG_BLOCK, &mask, &oldmask);

    for (int i=1;i<argc;i++) {
        if (!strcmp("--help", argv[i])) {
            usage(argv[0]);
            break;
        } else if (!strcmp("--device", argv[i])) {
            if (device) {
                little_close(device);
            }
            devicenum = atoi(argv[++i]);
            if (devicenum < 0) {
                usage(argv[0]);
                myshutdown(-1);
            }
        } else if (!strcmp("--host", argv[i])) {
            host = argv[++i];
        } else if (!strcmp("--port", argv[i])) {
            port = atoi(argv[++i]);
        } else if (!strcmp("--quiet", argv[i])) {
            quiet = 1;
        } else if (!strcmp("list", argv[i])) {
            if (device) {
                little_close(device);
                device = NULL;
            }
            list();
        } else if (!strcmp("get", argv[i])) {
            if (!device) {
                device = open(devicenum);
            }
            printf("%d\n", little_get(device));
        } else if (!strcmp("poll", argv[i])) {
            uint8_t oldret = 99;
            if (!device) {
                device = open(devicenum);
                state = little_get(device);
            }
            while (device) {
                int8_t ret = little_get(device);
                if (ret < 0) {
                    exit(-1);
                }
                if (ret != oldret) {
                    printf("%d\n", ret);
                    oldret = ret;
                }
                usleep(100000);
            }
        } else if (!strcmp("reset", argv[i])) {
            if (device) {
                little_close(device);
                device = NULL;
            }
            little_resetall();
        } else if (!strcmp("on", argv[i])) {
            if (!device) {
                device = open(devicenum);
                state = little_get(device);
            }
            little_set(device, state |= 2);
        } else if (!strcmp("ledon", argv[i])) {
            if (!device) {
                device = open(devicenum);
                state = little_get(device);
            }
            little_set(device, state |= 4);
        } else if (!strcmp("off", argv[i])) {
            if (!device) {
                device = open(devicenum);
                state = little_get(device);
            }
            little_set(device, state &= ~2);
        } else if (!strcmp("ledoff", argv[i])) {
            if (!device) {
                device = open(devicenum);
                state = little_get(device);
            }
            little_set(device, state &= ~4);
        } else if (!strcmp("scratch", argv[i])) {
            scratch(host, port);
            error = -1;
        } else {
            usage(argv[0]);
            error = -1;
            break;
        }
    }
    exit(error);
}
